@@ -28,7 +28,9 @@ namespace eval ::pki {
 	variable oids
 	array set oids {
 		1.2.840.113549.1.1.1           rsaEncryption
+		1.2.840.113549.1.1.4           md5WithRSAEncryption
 		1.2.840.113549.1.1.5           sha1WithRSAEncryption
+		1.2.840.113549.1.1.11          sha256WithRSAEncryption
 		1.2.840.113549.2.5             md5
 		1.3.14.3.2.26                  sha1
 		2.16.840.1.101.3.4.2.1         sha256
@@ -187,7 +189,7 @@ namespace eval ::pki {
 
 	variable handlers
 	array set handlers {
-		rsa                            {::pki::rsa::encrypt ::pki::rsa::decrypt ::pki::rsa::generate ::pki::rsa::serialize_key}
+		rsa                            {::pki::rsa::encrypt ::pki::rsa::decrypt ::pki::rsa::generate ::pki::rsa::serialize_key ::pki::rsa::serialize_public_key}
 	}
 
 	variable INT_MAX [expr {[format "%u" -1] / 2}]
@@ -426,6 +428,34 @@ proc ::pki::rsa::decrypt {mode input keylist} {
 	return $retval
 }
 
+proc ::pki::rsa::serialize_public_key {keylist} {
+	array set key $keylist
+
+	foreach entry [list n e] {
+		if {![info exists key($entry)]} {
+			return -code error "Key does not contain an element $entry"
+		}
+	}
+
+	set pubkey [::asn::asnSequence \
+		[::asn::asnBigInteger [::math::bignum::fromstr $key(n)]] \
+		[::asn::asnBigInteger [::math::bignum::fromstr $key(e)]] \
+		]
+	set pubkey_algo_params [::asn::asnNull]
+
+	binary scan $pubkey B* pubkey_bitstring
+
+	set ret [::asn::asnSequence \
+		[::asn::asnSequence \
+				[::asn::asnObjectIdentifier [::pki::_oid_name_to_number rsaEncryption]] \
+				$pubkey_algo_params \
+			] \
+			[::asn::asnBitString $pubkey_bitstring] \
+			]
+
+	return [list data $ret begin "-----BEGIN PUBLIC KEY-----" end "-----END PUBLIC KEY-----"]
+}
+
 proc ::pki::rsa::serialize_key {keylist} {
 	array set key $keylist
 
@@ -480,6 +510,9 @@ proc ::pki::_lookup_command {action keylist} {
 		"serialize_key" {
 			set idx 3
 		}
+		"serialize_public_key" {
+			set idx 4
+		}
 	}
 
 	set cmdlist $::pki::handlers($type)
@@ -743,6 +776,62 @@ proc ::pki::key {keylist {password ""} {encodePem 1}} {
 	return $retval
 }
 
+proc ::pki::public_key {keylist {password ""} {encodePem 1}} {
+	set serialize_key [::pki::_lookup_command serialize_public_key $keylist]
+
+	if {$serialize_key eq ""} {
+		array set key $keylist
+
+		return -code error "Do not know how to serialize an $key(type) key"
+	}
+
+	array set retval_parts [$serialize_key $keylist]
+
+	if {$encodePem} {
+		set retval [::pki::_encode_pem $retval_parts(data) $retval_parts(begin) $retval_parts(end) $password]
+	} else {
+		if {$password != ""} {
+			return -code error "DER encoded keys may not be password protected"
+		}
+
+		set retval $retval_parts(data)
+	}
+
+	return $retval
+}
+
+proc ::pki::parse {text {errorOnUnknownType 0}} {
+	set rc [list]
+	while {[regexp {^.*?-----BEGIN (.*?)-----(.*?)-----END (.*?)-----(.*)$} $text - type body type2 text]} {
+		if {$type != $type2} {
+			return -code error "BEGIN and END types do not match ($type and $type2)"
+		}
+		set body "-----BEGIN $type-----\n$body\n-----END $type-----\n"
+
+		switch -- $type {
+			"RSA PRIVATE KEY" {
+				lappend rc key [::pki::pkcs::parse_key $body]
+			}
+			"PUBLIC KEY" {
+				lappend rc public_key [::pki::pkcs::parse_public_key $body]
+			}
+			"CERTIFICATE" {
+				lappend rc certificate [::pki::x509::parse_cert $body]
+			}
+			"CERTIFICATE REQUEST" {
+				lappend rc certificate_request [::pki::pkcs::parse_csr $body]
+			}
+			default {
+				if {$errorOnUnknownType} {
+					return -code error "Unable to parse key with type $type"
+				}
+			}
+		}
+	}
+
+	return $rc
+}
+
 proc ::pki::_parse_init {} {
 	if {[info exists ::pki::_parse_init_done]} {
 		return
@@ -955,6 +1044,39 @@ proc ::pki::_parse_pem {pem begin end {password ""}} {
 	return [array get ret]
 }
 
+proc ::pki::pkcs::parse_public_key {key {password ""}} {
+	array set parsed_key [::pki::_parse_pem $key "-----BEGIN PUBLIC KEY-----" "-----END PUBLIC KEY-----" $password]
+
+	set key_seq $parsed_key(data)
+
+	::asn::asnGetSequence key_seq pubkeyinfo
+		::asn::asnGetSequence pubkeyinfo pubkey_algoid
+			::asn::asnGetObjectIdentifier pubkey_algoid oid
+		::asn::asnGetBitString pubkeyinfo pubkey
+	set ret(pubkey_algo) [::pki::_oid_number_to_name $oid]
+
+	switch -- $ret(pubkey_algo) {
+		"rsaEncryption" {
+			set pubkey [binary format B* $pubkey]
+
+			::asn::asnGetSequence pubkey pubkey_parts
+				::asn::asnGetBigInteger pubkey_parts ret(n)
+				::asn::asnGetBigInteger pubkey_parts ret(e)
+
+			set ret(n) [::math::bignum::tostr $ret(n)]
+			set ret(e) [::math::bignum::tostr $ret(e)]
+			set ret(l) [expr {int([::pki::_bits $ret(n)] / 8.0000 + 0.5) * 8}]
+			set ret(type) rsa
+		}
+		default {
+			error "Unknown algorithm"
+		}
+	}
+
+	return [array get ret]
+}
+
+
 proc ::pki::pkcs::parse_key {key {password ""}} {
 	array set parsed_key [::pki::_parse_pem $key "-----BEGIN RSA PRIVATE KEY-----" "-----END RSA PRIVATE KEY-----" $password]
 
@@ -1130,7 +1252,7 @@ proc ::pki::x509::parse_cert {cert} {
 					set ext_value [list $ext_critical]
 
 					switch -- $ext_oid {
-						id-ce-basicConstraints {
+						"id-ce-basicConstraints" {
 							::asn::asnGetSequence ext_value_seq ext_value_bin
 
 							if {$ext_value_bin != ""} {
@@ -1266,7 +1388,7 @@ proc ::pki::x509::validate_cert {cert args} {
 		set critical [lindex $ext_val 0]
 
 		switch -- $ext_id {
-			id-ce-basicConstraints {
+			"id-ce-basicConstraints" {
 				set CA [lindex $ext_val 1]
 				set CAdepth [lindex $ext_val 2]
 			}
@@ -1562,6 +1684,27 @@ proc ::pki::x509::create_cert {signreqlist cakeylist serial_number notBefore not
 						set extvalue [::asn::asnSequence [::asn::asnBoolean $allowCA] [::asn::asnInteger $caDepth]]
 					}
 				}
+				"id-ce-subjectAltName" {
+					set critical [lindex $extvalue 0]
+	
+					unset -nocomplain altnames
+
+					foreach {altnametype altnamevalue} [lrange $extvalue 1 end] {
+						switch -- [string tolower $altnametype] {
+							"rfc822name" {
+								lappend altnames [::asn::asnChoice 1 [::asn::asnIA5String $altnamevalue]]
+							}
+							"dnsname" {
+								lappend altnames [::asn::asnChoice 2 [::asn::asnIA5String $altnamevalue]]
+							}
+							default {
+								return -code error "Unknown subjectAltName type: $altnametype"
+							}
+						}
+					}
+
+					set extvalue [::asn::asnSequence {*}$altnames]
+				}
 				default {
 					return -code error "Unknown extension: $extension"
 				}
@@ -1881,4 +2024,4 @@ proc ::pki::rsa::generate {bitlength {exponent 0x10001}} {
 # # ## ### ##### ######## #############
 ## Ready
 
-package provide pki 0.5
+package provide pki 0.7
